/*
 * main.c
 *
 *  Created on: 20.7.2010
 *      Author: M.inko
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "avrlibdefs.h"
#include "avrlibtypes.h"
#include "timer.h"


#define CONNECTION_IR	//connection by IR LED; undef if connected by direct wire

#define PORT_OC0A	PORTB			//PORT, DDR, and BIT for OC0A output
#define DDR_OC0A	DDRB
#define BIT_OC0A	0

#define ADC_PRESCALER	(0|BV(ADPS2)/*|BV(ADPS1)|BV(ADPS0)*/)		//prescaler bits to get 50-200kHz clock for ADC; 100b => 16.. F_CPU=1.2MHz=>ADCclk=75kHz
#define ADC_MUX			(0|BV(MUX1)/*|BV(MUX0)*/)	//channel 2:  MUX1 MUX0 = 10
#define	ADC_DIGITAL_DISABLE_PIN	ADC2D	//disable digital function of ADC2


#define CARRIER_INTERVAL	550		//kenwood reacts with values 300 - 800

#ifdef CONNECTION_IR
	//value for register OCR0A in 8-bit timer0, to get proper carrier freq. formula is Fosc/(2.N.(1+OCR0A)) = Fcarrier; N(prescaler)=1
	#define CARRIER_FREQ		((double)38000)	//value is not critical
	#define OCR0A_FOR_CARRIER	((BYTE)(((double)F_CPU/(2.0*CARRIER_FREQ))-1.0))

	#define CARRIER 	DDR_OC0A |=  BV(BIT_OC0A);		//output for OC0A
	#define NO_CARRIER 	DDR_OC0A &= ~BV(BIT_OC0A);		//input for OC0A
#else
	#define CARRIER 	sbi(PORT_OC0A, BIT_OC0A);		//H on OC0A
	#define NO_CARRIER 	cbi(PORT_OC0A, BIT_OC0A);		//L on OC0A
#endif

// commands marked with * works on KDC-W4537
#define CMD_0			0xb900 //   preset (memory) 0
#define CMD_1			0xb901 // * preset (memory) 1
#define CMD_2			0xb902 // * preset (memory) 2
#define CMD_3			0xb903 // * preset (memory) 3
#define CMD_4			0xb904 // * preset (memory) 4
#define CMD_5			0xb905 // * preset (memory) 5
#define CMD_6			0xb906 // * preset (memory) 6
#define CMD_7			0xb907 //   preset (memory) 7
#define CMD_8			0xb908 //   preset (memory) 8
#define CMD_9			0xb909 //   preset (memory) 9

							   //simulation of joystick, reacts also in menu
#define CMD_LEFT		0xb90a // * (TRACK_PREV) tuner: serch<< / CD,USB: 1st CMD moves track to start, 2nd CMD moves to previous track
#define CMD_RIGHT		0xb90b // * (TRACK_NEXT) tuner: serch>> / CD,USB: moves to next track
#define CMD_DOWN		0xb90c // * (REWIND)	 tuner: switch to MW / CD: nothing / USB: move to 1st track
#define CMD_UP			0xb90d // * (FAST_FORW)	 tuner: switch between FM1, FM2, FM3 / CD: nothing / USB: move to 1st track
#define CMD_PUSH		0xb917 // * switch between settings Equalizer scheme (natural, jazz,...), BASS, TREBLE, BALANCE, FADER

#define CMD_PLAY_PAUSE	0xb90e // * tuner: switch on disp PTY / CD,USB: play/pause

#define CMD_MENU		0xb912 // * enter/exit the menu
#define CMD_SOURCE		0xb913 // * switches sources stand by, tuner, USB, CD, AUX
#define CMD_VOLUME_UP	0xb914 // * volume up
#define CMD_VOLUME_DOWN	0xb915 // * volume down
#define CMD_ATT			0xb916 // * switch ATT on/off

#define CMD_TUNER		0xb91c // * switch tuner/stand by
#define CMD_TAPE		0xb91d //   switch tape
#define CMD_CD_USB		0xb91e // * switch on CD/USB (if CD not present switch on USB directly)
#define CMD_AUX			0xb91f // * switch on AUX

#define CMD_DNPP		0xb95e //   ?


#define DECISIONS	8
PROGMEM BYTE decisions[DECISIONS] = {241, 210, 177, 146, 115, 82, 53, 20};
PROGMEM WORD commands[DECISIONS+1] = {0, 0, CMD_1, CMD_LEFT, CMD_RIGHT, CMD_ATT, CMD_VOLUME_UP, CMD_VOLUME_DOWN};
#define CMD_INDEX_VOLUME_UP		6
#define CMD_INDEX_VOLUME_DOWN 	7



#define JITTER		10		//number of ADC measurement loops with the same result to recognize that the key is correctly pressed
							//together with ACD frequency (set by ADC prescaler) affetcs autorepeat delay after 1st press and next cadence

#define AUTOREP_CNT	20		//number of command recognition loops between 1st cmd send and the next sending of the same cmd
							//then the sending is autorepeated quickly
/*************************************************************************************************/
void vSendByte(BYTE b)
{
BYTE bMask=1;
BYTE i;
	for(i=0; i<8; i++)
	{
		NO_CARRIER
		DELAY_US(CARRIER_INTERVAL);
		if(b&bMask)	DELAY_US(2*CARRIER_INTERVAL); //bit==1
		CARRIER
		DELAY_US(CARRIER_INTERVAL);
		bMask=bMask<<1;
	}
}
/*************************************************************************************************/
void vSendCommand(WORD cmd)
{
	//start seq
	CARRIER
	DELAY_US(18*CARRIER_INTERVAL);
	NO_CARRIER
	DELAY_US(9*CARRIER_INTERVAL);
	CARRIER
	DELAY_US(CARRIER_INTERVAL);
	//code
	vSendByte((BYTE)(cmd>>8));
	vSendByte((BYTE)(~(cmd>>8)));
	vSendByte((BYTE)(cmd&0xFF));
	vSendByte((BYTE)(~(cmd&0xFF)));
	//stop
	NO_CARRIER
}
/*************************************************************************************************/
int	main(void)
{
BYTE adc, i, j, cmdindex=0, cmdindexold=0,  cmdjitter=0, send, autorep_pause_cnt=0;

	wdt_enable(WDTO_2S);

	//set pull-ups on all ports as datasheet recommends
								//except of OC0A
								//NPN must be defaultly closed
	PORT_OC0A = ~BV(BIT_OC0A);	//default value of the OC0A is 0 (pull-up disconnected if input or 0V if output)


#ifdef CONNECTION_IR
	//default direction of the OC0A is input (it is default of DDR)

	//CTC mode: WGM02(TCCR0B),WGM01(TCCR0A),WGM00(TCCR0A) = 010
	//Toggle OC0A on Compare Match: COM0A1(TCCR0A),COM0A0(TCCR0A) = 01
	outb(TCCR0A, BV(WGM01) | BV(COM0A0));

	outb(OCR0A, OCR0A_FOR_CARRIER);

	// run timer on full freq
	timer0SetPrescaler(TIMER_CLK_DIV1);
#else
	//set OC0A as output (with value 0)
	DDR_OC0A = BV(BIT_OC0A);
#endif

	//enable ADC and set the prescaler
	outb(ADCSRA, BV(ADEN) | ADC_PRESCALER);

	//reference voltage is VCC: REFS0 = 0
	//left adjusted (it is enough to read high byte for 8-bit precision) ADLAR=1
	//set MUX bits
	outb(ADMUX, BV(ADLAR) | ADC_MUX);

	//disable digital input function on used channel pin
	outb(DIDR0, BV(ADC_DIGITAL_DISABLE_PIN));

	for(;;)	/* main loop */
	{
		for(j=0; j<JITTER; j++)
		{
			wdt_reset();
			sbi(ADCSRA, ADSC);				//start conversion
			while(inb(ADCSRA) & BV(ADSC));	//wait for finishing of conversion
			adc = inb(ADCH);				//get 8-bit result

			cmdindex=0;
			for(i=0; i<DECISIONS; i++)			//decide which button is pressed
			{
				if(adc > pgm_read_byte(decisions + i))
				{
					if(i==0) break; 	//steering wheel buttons are not connected
					if(i==1)			//not pressed, e.g. released!
					{
						cmdindexold = 0;		//reset last command
						autorep_pause_cnt = 0;	//reset pause counter for new command
						break;
					}
					cmdindex = i;				//key press detected
					break;
				}
			}
			if(j==0) cmdjitter = cmdindex;		//1st keypress detection in jitter validation, only store
			else if(cmdjitter != cmdindex)		//each next jitter validation loop, if jitter error detected (cmdindex is different than last one)
			{
				cmdindex=0;						//mark as not pressed
				break;							//break from jitter loop - next jitter loop can start immediately
			}									//for faster recognition of correct keypress
		}
		if(cmdindex)					//some key is correctly pressed without jitter error
		{
			send=0;
			if(cmdindex != cmdindexold)	//new command (1st immediatelly key press) - always send
			{
				send=1;
			}
			else if(cmdindex==CMD_INDEX_VOLUME_DOWN || cmdindex==CMD_INDEX_VOLUME_UP)//the same command as previous, test for autorepeat on volume only
			{
				if(autorep_pause_cnt < AUTOREP_CNT)	//pause for proper number of loops
				{
					DELAY_US(10000);
					autorep_pause_cnt++;	//only increment counter
				}
				else								//counter reached, send command and do NOT update counter
					send=1;
			}
			if(send)	//send condition fulfilled
			{
				vSendCommand(pgm_read_word(commands + cmdindex));
				cmdindexold = cmdindex;			//store current command
				DELAY_US(70000);
			}
		}
	}
	return 0;
}

/* ------------------------------------------------------------------------- */
